import 'dart:async';
import 'dart:io';

import 'package:firebase_storage/firebase_storage.dart';
import 'package:flutter/foundation.dart';
import 'package:mime/mime.dart';
import 'package:_private_chat/_private_chat.dart';
import 'package:path/path.dart' as p;
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:get/get.dart';
import 'package:_private_core_network/network_resources/resources.dart';
import 'package:_private_chat/utils/app_uploader.dart';
import 'package:_private_chat/utils/utils.dart';
import 'package:_private_core/_private_core.dart';
import 'package:_private_core/widgets/widgets.dart';

import '../asset/bloc/chat_detail_asset_bloc.dart';
import '../chat_helper.dart';
import '../message.dart';
import '../widgets/widget_message_content.dart';

part 'chat_detail_event.dart';
part 'chat_detail_state.dart';

class ChatDetailBloc extends Bloc<ChatDetailEvent, ChatDetailState> {
  final SeenState seenState = SeenState(<String, int>{});
  final ReactionState reactionState = ReactionState(<dynamic, dynamic>{});

  final ValueNotifier<Map> progressUpload = ValueNotifier({});
  final ValueNotifier<Map<String, dynamic>> messageUpdatedValue =
      ValueNotifier({});
  late ChatDetailParams params;
  late ScrollController _scrollController;
  late TickerProvider _tickerProvider;
  late GlobalKey targetAssistantKey = GlobalKey();
  StreamSubscription? _messagesSubscription;
  StreamSubscription? _conversationSubscription;

  final TextEditingController inputController = TextEditingController();
  String get textSelectedInTextField =>
      inputController.selection.textInside(inputController.text);
  final FocusNode inputFocus = FocusNode();

  Map? conversationData;

  ChatConversationType get conversationType => params.conversationType;

  String get conversationId => conversationData?[DB_KEY_CONVERSATION_ID] ?? '';

  String get peerId => params.conversationType == ChatConversationType.solo
      ? params.peerId ??
          (conversationData?[DB_KEY_MEMBERS]
              .firstWhere((e) => e != loggedFirebaseId)) ??
          ''
      : '';

  String get peerAvatar {
    Map? user = Get.find<ChatListBloc>().state.usersByKeys[peerId];
    return user?['thumbnail'] ?? params.thumbnail ?? '';
  }

  String get peerName {
    Map? user = Get.find<ChatListBloc>().state.usersByKeys[peerId];
    return user?['displayName'] ?? user?['displayName'] ?? params.name ?? '';
  }

  String get peerUniqueId {
    Map? user = Get.find<ChatListBloc>().state.usersByKeys[peerId];
    return user?['uniqueId'] ?? '';
  }

  DocumentReference<Map<String, dynamic>> get _docRef =>
      FirebaseFirestore.instance
          .collection(C_CONVERSATIONS)
          .doc(conversationId);

  CollectionReference<Map<String, dynamic>> get _colRef =>
      _docRef.collection(C_MESSAGES);

  Future<Map?> getNextMessage(MessageType messageType, timestamp) async {
    QuerySnapshot q = await _colRef
        .where(DB_KEY_MESSAGE_TYPE, isEqualTo: messageType.index)
        .where(DB_KEY_TIMESTAMP, isGreaterThan: timestamp)
        .limit(1)
        .get();
    if (q.size != 0) {
      return q.docs.single.data() as Map;
    }
  }

  ChatDetailBloc() : super(ChatDetailState()) {
    on<ChatDetailInitialEvent>(_onInitial);
    on<ChatDetailDisposeEvent>(_onDispose);
    on<ChatDetailLoadMoreMessagesEvent>(_onLoadMoreMessages);
    on<ChatDeleteMessagesEvent>(_onDeleteMessage);
    on<ChatUpdateMessagesEvent>(_onUpdateMessages);
    on<ChatSendMessageEvent>(_onSendMessage);
    on<ChatDetailReplyMessagesEvent>(_onReplyMessage);
    on<ChatDetailEditMessagesEvent>(_onEditMessage);
    on<ChatDetailEditLastMessagesEvent>(_onEditLastMessage);
    on<ChatDetailResetEditOrReplyEvent>(_onResetState);
    on<ChatDetailSendReactionMessagesEvent>(_onReaction);
    on<ChatDetailUnsendReactionMessagesEvent>(_onRemoveReaction);
    on<ChatDetailCopyMessageEvent>(_onCopyMessage);
    on<ChatDetailPastTextInputEvent>(_onPassText);
  }

  _onPassText(event, Emitter<ChatDetailState> emit) {
    if (event.text != null) {
      inputController.text += event.text;
      inputFocus.requestFocus();
      inputController.selection = TextSelection.fromPosition(
          TextPosition(offset: inputController.text.length));
    }
  }

  _onCopyMessage(event, Emitter<ChatDetailState> emit) {
    Clipboard.setData(ClipboardData(
        text: ChatHelper.decrypt(
            conversationData![DB_KEY_ENCRYPT_KEY], event.text)));
  }

  _onRemoveReaction(ChatDetailUnsendReactionMessagesEvent event,
      Emitter<ChatDetailState> emit) async {
    Map<String, dynamic> data = (event.doc[DB_KEY_REACTIONS] ?? {});
    data.remove(loggedFirebaseId);
    _colRef
        .doc('${event.doc[DB_KEY_TIMESTAMP]}')
        .set({DB_KEY_REACTIONS: data}, SetOptions(merge: true));
  }

  _onReaction(ChatDetailSendReactionMessagesEvent event,
      Emitter<ChatDetailState> emit) async {
    Map<String, dynamic> data = (event.doc[DB_KEY_REACTIONS] ?? {});
    data.addAll({loggedFirebaseId: event.value});
    _colRef
        .doc('${event.doc[DB_KEY_TIMESTAMP]}')
        .set({DB_KEY_REACTIONS: data}, SetOptions(merge: true));
  }

  _onResetState(ChatDetailResetEditOrReplyEvent event,
      Emitter<ChatDetailState> emit) async {
    emit(state.update(isEditKeyboard: false, isReplyKeyboard: false));
  }

  _onEditLastMessage(event, Emitter<ChatDetailState> emit) async {
    if (state.messages != null &&
        state.messages!.any((e) =>
            (e.tempDoc != null &&
                e.tempDoc![DB_KEY_FROM] == loggedFirebaseId) ||
            (e.doc != null && e.doc![DB_KEY_FROM] == loggedFirebaseId))) {
      var message = state.messages!.lastWhere((e) =>
          (e.tempDoc != null && e.tempDoc![DB_KEY_FROM] == loggedFirebaseId) ||
          (e.doc != null && e.doc![DB_KEY_FROM] == loggedFirebaseId));
      inputFocus.unfocus();
      Map<String, dynamic> editDoc =
          message.tempDoc ?? (message.doc!.data() as Map<String, dynamic>);
      Timer(const Duration(milliseconds: 500), () {
        add(ChatDetailEditMessagesEvent(editDoc));
      });
    }
  }

  _onEditMessage(
      ChatDetailEditMessagesEvent event, Emitter<ChatDetailState> emit) async {
    inputController.text = ChatHelper.decrypt(
        conversationData![DB_KEY_ENCRYPT_KEY], event.editDoc[DB_KEY_CONTENT]);
    inputFocus.requestFocus();
    inputController.selection = TextSelection.fromPosition(
        TextPosition(offset: inputController.text.length));
    emit(state.update(
        editDoc: event.editDoc, isEditKeyboard: true, isReplyKeyboard: false));
  }

  _onReplyMessage(
      ChatDetailReplyMessagesEvent event, Emitter<ChatDetailState> emit) async {
    inputFocus.requestFocus();
    emit(state.update(
        replyDoc: event.repllyDoc,
        isEditKeyboard: false,
        isReplyKeyboard: true));
  }

  _onUploadProgress(
      String fileName, MessageType type, TaskSnapshot taskSnapshot) async {
    switch (taskSnapshot.state) {
      case TaskState.running:
        final progress =
            100.0 * (taskSnapshot.bytesTransferred / taskSnapshot.totalBytes);
        progressUpload.value = {
          'fileName': fileName,
          'messageType': type.index,
          'progress': progress.toInt()
        };
        appDebugPrint("Upload is $progress% complete.");
        break;
      case TaskState.paused:
        appDebugPrint("Upload is paused.");
        break;
      case TaskState.canceled:
        appDebugPrint("Upload was canceled");
        progressUpload.value = {};
        break;
      case TaskState.error:
        appDebugPrint("Upload was error");
        progressUpload.value = {};
        break;
      case TaskState.success:
        appDebugPrint("Upload was success");
        Timer(const Duration(milliseconds: 200), () {
          progressUpload.value = {};
        });
        break;
      default:
        Timer(const Duration(milliseconds: 200), () {
          progressUpload.value = {};
        });
        break;
    }
  }

  String _tempMessageContent(MessageType messageType, text, sender) {
    switch (messageType) {
      case MessageType.gif:
      case MessageType.image:
        return '$sender: an image';
      case MessageType.audio:
        return '$sender: an audio';
      case MessageType.lottie:
        return '$sender: a smile';
      default:
        return '$sender: $text';
    }
  }

  _onSendMessage(
      ChatSendMessageEvent event, Emitter<ChatDetailState> emit) async {
    int timestamp = DateTime.now().millisecondsSinceEpoch;
    String? assetUrl;
    String? assetExt;
    String? assetFileName;
    String? assetFileMime;
    int? assetFileDuration;
    if (event.type == MessageType.gif) {
      assetExt = 'gif';
      assetFileName = '';
      assetFileMime = 'image/gif';
      assetUrl = event.text;
    } else if (event.type == MessageType.audio) {
      assetExt = event.imageExt;
      assetFileName = event.imageFileName;
      assetFileMime = event.imageFileMime;
      assetFileDuration = event.assetFileDuration;
      assetUrl = await AppUploader.uploadFile(
        data: event.imageData,
        path: 'conversations/$conversationId',
        userId: loggedFirebaseId,
        meta: SettableMetadata(contentType: assetFileMime),
        onUploading: (task) {
          _onUploadProgress(assetFileName ?? '', event.type, task);
        },
      );
    } else if (event.type == MessageType.image) {
      if (event.imageData != null) {
        assetExt = event.imageExt;
        assetFileName = event.imageFileName;
        assetFileMime = event.imageFileMime;
        assetUrl = await AppUploader.uploadFile(
          data: event.imageData,
          path: 'conversations/$conversationId',
          userId: loggedFirebaseId,
          meta: SettableMetadata(contentType: assetFileMime),
          onUploading: (task) {
            _onUploadProgress(assetFileName ?? '', event.type, task);
          },
        );
      } else {
        // File? image = event.image ??
        //     await chatOptions!.takeFile!(
        //         fileType: PTakeFileType.single,
        //         fileTypeAllow: PTakeFileTypeAllow.image);
        // if (image == null) return;
        // assetExt = p.extension(image.path);
        // assetFileName = p.basename(image.path);
        // assetFileMime = lookupMimeType(image.path);
        // assetUrl = await AppUploader.uploadFile(
        //   file: image,
        //   path: 'conversations/$conversationId',
        //   meta: SettableMetadata(contentType: assetFileMime),
        //   userId: loggedFirebaseId,
        //   onUploading: (task) {
        //     _onUploadProgress(assetFileName ?? '', event.type, task);
        //   },
        // );
      }
    } else if (state.isEditKeyboard) {
      String content =
          ChatHelper.encrypt(conversationData![DB_KEY_ENCRYPT_KEY], event.text);
      _colRef.doc('${state.editDoc![DB_KEY_TIMESTAMP]}').update({
        DB_KEY_IS_EDITTED: true,
        DB_KEY_CONTENT: content,
      });
      add(ChatDetailResetEditOrReplyEvent());
      return;
    }

    String content =
        ChatHelper.encrypt(conversationData![DB_KEY_ENCRYPT_KEY], event.text);
    Map<String, dynamic> data = {
      DB_KEY_CHATID: conversationId,
      DB_KEY_FROM: loggedFirebaseId,
      DB_KEY_SENDER: params.conversationType == ChatConversationType.solo
          ? appPrefs.loginUser!.fullName
          : params.name,
      DB_KEY_TIMESTAMP: timestamp,
      DB_KEY_CONTENT: content,
      DB_KEY_MESSAGE_TYPE: event.type.index,
      DB_KEY_IS_REPLY: state.isReplyKeyboard && event.type != MessageType.image,
      DB_KEY_REPLY_DOC: state.replyDoc,
      DB_KEY_BODY_NOTIF: _tempMessageContent(
          event.type, event.text, appPrefs.loginUser!.fullName),
      DB_KEY_PLATFORM_ID: valuePlatformByConfig(),
    };
    if (assetUrl != null) {
      data.addAll({
        DB_KEY_CONTENT_URL: assetUrl,
        DB_KEY_CONTENT_FILE_INFOS: {
          DB_KEY_FILE_EXT: assetExt,
          DB_KEY_FILE_FILE_NAME: assetFileName,
          DB_KEY_FILE_MIME: assetFileMime,
          DB_KEY_FILE_DURATION: assetFileDuration,
        },
      });
    }
    await ChatHelper.setLastMessageTime(conversationData, timestamp,
        alsoMembersDetail: true);
    Future messaging =
        _colRef.doc('$timestamp').set(data, SetOptions(merge: true)).then((e) {
      _colRef.doc('$timestamp').update({DB_KEY_BODY_NOTIF: ''});
      _refreshListChatEachChange();
    });

    ChatHelper.addMessage(peerId, timestamp, messaging);

    List<Message> messages = state.messages ?? [];

    Message m = Message(
      tempDoc: data,
      animationController: AnimationController(
        duration: messageAnimationDuration,
        reverseDuration: messageReverseAnimationDuration,
        vsync: _tickerProvider,
      ),
      snappableKey: GlobalKey<SnappableState>(),
      timestamp: timestamp,
      widgetMessageOwner: (messages) => _builderProvider(
        timestamp: timestamp,
        child: WidgetMessageOwner(
          messages: messages,
          data: data,
          peerId: data[DB_KEY_FROM],
          toPeerId: peerId,
          delivered: ChatHelper.getMessageStatus(peerId, timestamp),
        ),
      ),
    )..animationController.forward();

    messages.add(m);

    emit(state.update(messages: messages));
    add(ChatDetailResetEditOrReplyEvent());
  }

  _onDispose(
      ChatDetailDisposeEvent event, Emitter<ChatDetailState> emit) async {
    Get.find<ChatDetailAssetBloc>().add(DisposeChatDetailAssetEvent());
    _messagesSubscription?.cancel();
    _conversationSubscription?.cancel();
    if (_typing == true) {
      Get.find<ConnectionStatusBloc>().add(ConnectionStatusSetStatusEvent(
          conversationStatus: ConversationStatus.none));
    }
    await ChatHelper.setLastTimeTimestamp(conversationData);
    Get.find<ChatListBloc>().add(ChatListInitEvent());
  }

  _onUpdateMessages(
      ChatUpdateMessagesEvent event, Emitter<ChatDetailState> emit) {
    emit(state.update());
  }

  _onDeleteMessage(
      ChatDeleteMessagesEvent event, Emitter<ChatDetailState> emit) {
    _deleteMessage(event.timestamp);
    emit(state.update());
  }

  bool _typing = false;
  _onInitial(
      ChatDetailInitialEvent event, Emitter<ChatDetailState> emit) async {
    if (!kIsWeb) {
      inputController.text = appPrefs.getDraftMessage(conversationId);
    }
    inputController.addListener(() {
      if (inputController.text.isNotEmpty && _typing == false) {
        Get.find<ConnectionStatusBloc>().add(ConnectionStatusSetStatusEvent(
            conversationStatus: ConversationStatus.typing,
            conversationId: conversationId));
        _typing = true;
      }
      if (inputController.text.isEmpty && _typing == true) {
        Get.find<ConnectionStatusBloc>().add(ConnectionStatusSetStatusEvent(
            conversationStatus: ConversationStatus.none));
        _typing = false;
      }
    });
    conversationData = null;
    _tickerProvider = event.tickerProvider;
    emit(state.update(
        isLoading: true, isEditKeyboard: false, isReplyKeyboard: false));
    if (params.conversationId != null) {
      conversationData = await ChatHelper.findConversationById(
          conversationId: params.conversationId!,
          type: params.conversationType);
    }
    conversationData ??= params.conversationType == ChatConversationType.solo
        ? await ChatHelper.findOrCreateSoloConversation(
            params.peerId!, params.peerType!.name)
        : await ChatHelper.findOrCreateGroupConversation(
            conversationId: params.conversationId,
            peers: params.peers,
            peersDetail: params.peersDetail,
            groupName: params.name,
            thumbnail: params.thumbnail,
          );
    ChatHelper.setLastTimeTimestamp(conversationData);

    _conversationSubscription =
        _docRef.snapshots().listen((DocumentSnapshot d) {
      appDebugPrint('conversationData: $conversationData');
      conversationData = d.data() as Map?;
      List members = conversationData![DB_KEY_MEMBERS_DETAIL];
      Map<String, dynamic> _ = {};
      for (var e in members) {
        _.addAll({e[DB_KEY_FIREBASE_ID]: e[DB_KEY_TIMESTAMP] ?? 0});
      }
      seenState.value = _;
    });

    List<Message> messages = <Message>[];
    await _colRef
        .orderBy(DB_KEY_TIMESTAMP)
        .limitToLast(LIMIT_MESSAGES_PERPAGE)
        .get()
        .then((docs) async {
      bool isEmpty = !docs.docs.isNotEmpty;
      if (docs.docs.isNotEmpty) {
        docs.docs.forEach((doc) {
          Map<String, dynamic> _doc = Map.from(doc.data());
          int? timestamp = _doc[DB_KEY_TIMESTAMP];
          Message m = Message(
              doc: doc,
              animationController: AnimationController(
                value: 1,
                duration: messageAnimationDuration,
                reverseDuration: messageReverseAnimationDuration,
                vsync: _tickerProvider,
              ),
              snappableKey: GlobalKey<SnappableState>(),
              widgetMessageOwner: (messages) => _builderProvider(
                    timestamp: timestamp,
                    child: WidgetMessageOwner(
                      messages: messages,
                      data: _doc,
                      peerId: _doc[DB_KEY_FROM],
                      toPeerId: peerId,
                      delivered: ChatHelper.getMessageStatus(peerId, timestamp),
                    ),
                  ),
              timestamp: timestamp);
          messages.add(m);
          Map<dynamic, dynamic> _ = Map.from(reactionState.value);
          _.addAll({doc.id: _doc[DB_KEY_REACTIONS]});
          reactionState.value = _;
        });
      }
      emit(state.update(messages: messages, isLoading: false));
      _isLoadingMore = false;

      Get.find<ChatDetailAssetBloc>()
          .add(InitChatDetailAssetEvent(conversationData));

      _messagesSubscription = _colRef.snapshots().listen((query) async {
        if (isEmpty ||
            query.docs.length != query.docChanges.length ||
            query.docChanges
                .where((doc) =>
                    doc.type == DocumentChangeType.modified ||
                    doc.type == DocumentChangeType.removed)
                .isNotEmpty) {
          List<Message> messages = state.messages ?? [];
          //----below action triggers when peer new message arrives
          query.docChanges.where((doc) {
            return doc.oldIndex <= doc.newIndex &&
                doc.type == DocumentChangeType.added &&
                (doc.doc.data() as Map)[DB_KEY_FROM] != loggedFirebaseId;
          }).forEach((change) {
            Map<String, dynamic> _doc = Map.from(change.doc.data()!);
            appDebugPrint('DocumentChangeType.received: $_doc');
            int? timestamp = _doc[DB_KEY_TIMESTAMP];
            Message m = Message(
              doc: change.doc,
              animationController: AnimationController(
                duration: messageAnimationDuration,
                reverseDuration: messageReverseAnimationDuration,
                vsync: _tickerProvider,
              ),
              snappableKey: GlobalKey<SnappableState>(),
              timestamp: timestamp,
              widgetMessageOwner: (messages) => _builderProvider(
                timestamp: timestamp,
                child: WidgetMessageOwner(
                  messages: messages,
                  data: _doc,
                  peerId: _doc[DB_KEY_FROM],
                  toPeerId: peerId,
                  delivered: ChatHelper.getMessageStatus(peerId, timestamp),
                ),
              ),
            )..animationController.forward();
            messages.add(m);
            ChatHelper.setLastTimeTimestamp(conversationData);
            _refreshListChatEachChange();
          });
          //----below action triggers when peer message get deleted
          query.docChanges
              .where((doc) => doc.type == DocumentChangeType.removed)
              .forEach((change) {
            Map<String, dynamic> _doc = Map.from(change.doc.data()!);
            appDebugPrint('DocumentChangeType.removed: $_doc');
            if (messages.any((e) => e.timestamp == _doc[DB_KEY_TIMESTAMP])) {
              _deleteMessage(
                  messages
                      .firstWhere((e) => e.timestamp == _doc[DB_KEY_TIMESTAMP])
                      .timestamp,
                  isOnlyLocal: true);
            }
          });

          //----below action triggers when peer message get update: edit or reaction..
          query.docChanges
              .where((doc) => doc.type == DocumentChangeType.modified)
              .forEach((change) {
            Map<String, dynamic> data = Map.from(change.doc.data()!);
            // appDebugPrint('DocumentChangeType.modified: $_doc');

            //Reactions
            Map<dynamic, dynamic> _ = Map.from(reactionState.value);
            _.addAll({change.doc.id: data[DB_KEY_REACTIONS]});
            reactionState.value = _;

            // Update real content
            Map<String, dynamic> y = Map.from(messageUpdatedValue.value);
            y.addAll({
              change.doc.id: ChatHelper.decrypt(
                  conversationData![DB_KEY_ENCRYPT_KEY], data[DB_KEY_CONTENT])
            });
            messageUpdatedValue.value = y;

            if (messages.any((e) => e.timestamp == data[DB_KEY_TIMESTAMP])) {
              messages
                  .firstWhere((e) => e.timestamp == data[DB_KEY_TIMESTAMP])
                  .tempDoc = null;
              messages
                  .firstWhere((e) => e.timestamp == data[DB_KEY_TIMESTAMP])
                  .doc = change.doc;
            }
          });

          isEmpty = messages.isEmpty;
          add(ChatUpdateMessagesEvent());
        }
      });

      _scrollController = event.scrollController;
      _scrollController.addListener(() {
        if (_scrollController.position.pixels >=
            _scrollController.position.maxScrollExtent) {
          add(ChatDetailLoadMoreMessagesEvent());
        }
      });
    });
  }

  bool _isLoadingMore = false;
  _onLoadMoreMessages(ChatDetailLoadMoreMessagesEvent event,
      Emitter<ChatDetailState> emit) async {
    if (_isLoadingMore) return;
    _isLoadingMore = true;
    _colRef
        .orderBy(DB_KEY_TIMESTAMP)
        .endBefore([state.messages!.first.timestamp])
        .get()
        .then((docs) {
          if (docs.docs.isNotEmpty) {
            List<QueryDocumentSnapshot<Map>> d = List.from(docs.docs);
            if (d.length > LIMIT_MESSAGES_PERPAGE) {
              d = d
                  .getRange(d.length - LIMIT_MESSAGES_PERPAGE, d.length)
                  .toList();
            }
            d.sort((a, b) {
              Map<String, dynamic> docA = Map.from(a.data());
              Map<String, dynamic> docB = Map.from(b.data());
              return docB[DB_KEY_TIMESTAMP] - docA[DB_KEY_TIMESTAMP];
            });
            List<Message> messages = List.from(state.messages ?? []);
            d.forEach((doc) {
              Map<String, dynamic> _doc = Map.from(doc.data());
              int? timestamp = _doc[DB_KEY_TIMESTAMP];
              Message m = Message(
                  doc: doc,
                  animationController: AnimationController(
                    value: 1,
                    duration: messageAnimationDuration,
                    reverseDuration: messageReverseAnimationDuration,
                    vsync: _tickerProvider,
                  ),
                  snappableKey: GlobalKey<SnappableState>(),
                  widgetMessageOwner: (messages) => _builderProvider(
                        timestamp: timestamp,
                        child: WidgetMessageOwner(
                          messages: messages,
                          data: _doc,
                          peerId: _doc[DB_KEY_FROM],
                          toPeerId: peerId,
                          delivered:
                              ChatHelper.getMessageStatus(peerId, timestamp),
                        ),
                      ),
                  timestamp: timestamp);
              Map<dynamic, dynamic> _ = Map.from(reactionState.value);
              _.addAll({doc.id: _doc[DB_KEY_REACTIONS]});
              reactionState.value = _;
              messages.insert(0, m);
            });
            state.messages = messages;
            add(ChatUpdateMessagesEvent());
            Timer(const Duration(seconds: 2), () {
              _isLoadingMore = false;
            });
          }
        });
  }

  _refreshListChatEachChange() {
    if (params.refreshListChatEachChange) {
      Get.find<ChatListBloc>().add(ChatListInitEvent());
    }
  }

  void _deleteMessage(timestamp, {bool isOnlyLocal = false}) {
    bool isExisting = state.messages!.any((msg) => msg.timestamp == timestamp);
    delete() {
      messageByTimestamp(timestamp).animationController.reverse();
      Timer(Duration(milliseconds: isExisting ? 500 : 100), () {
        List<Message> messages = List.from(state.messages ?? []);
        messages.removeWhere((msg) => msg.timestamp == timestamp);
        state.messages = messages;
        if (!isOnlyLocal) _colRef.doc('$timestamp').delete();
        ChatHelper.setLastMessageTime(
            conversationData, messages.last.timestamp);
      });
    }

    if (isExisting) {
      if (messageByTimestamp(timestamp).snappableKey.currentState != null) {
        messageByTimestamp(timestamp)
            .snappableKey
            .currentState
            ?.prepare()
            .then((value) {
          messageByTimestamp(timestamp)
              .snappableKey
              .currentState
              ?.snap()
              .then((value) {
            delete();
          });
        });
      } else {
        delete();
      }
    }
  }

  Message messageByTimestamp(timestamp) =>
      state.messages!.firstWhere((msg) => msg.timestamp == timestamp);

  Widget _builderProvider({timestamp, child}) {
    return SeenProvider(
      timestamp: timestamp.toString(),
      reaction: reactionState,
      seen: seenState,
      child: child,
    );
  }
}
