import 'dart:async';
import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:flutter_keyboard_visibility/flutter_keyboard_visibility.dart';

import 'package:get/get.dart' as getx;
import 'package:get/get.dart';
import 'package:_private_core/_private_core.dart';
import '../chat_list/bloc/chat_list_bloc.dart';
import 'bloc/chat_detail_bloc.dart';
import 'message.dart';
import 'widgets/widget_input.dart';
import 'widgets/widget_profile.dart';

enum ChatDetailType { users, pros }

enum ChatConversationType { solo, group }

class ChatDetailParams {
  ChatConversationType conversationType;
  String? conversationId;

  //Solo type
  String? peerId;
  ChatDetailType? peerType;

  //Group type
  List<String>? peers;
  List? peersDetail;

  //This will call get again data for list conversation
  bool refreshListChatEachChange;

  String? name;
  String? thumbnail;
  int? unread;

  ChatDetailParams(
    this.conversationType, {
    this.conversationId,
    this.peerType,
    this.peerId,
    this.peers,
    this.peersDetail,
    this.name,
    this.thumbnail,
    this.unread,
    this.refreshListChatEachChange = false,
  });
}

class ChatDetailScreen extends StatefulWidget {
  static double get padding => 16;

  final ChatDetailParams? params;
  final PChatOptions? options;

  const ChatDetailScreen({
    Key? key,
    this.params,
    this.options,
  }) : super(key: key);

  static String route = '/chat_detail';

  @override
  _ChatDetailScreenState createState() => _ChatDetailScreenState();
}

class _ChatDetailScreenState extends State<ChatDetailScreen>
    with TickerProviderStateMixin {
  final _scrollController = ScrollController();
  final _inputController = TextEditingController();
  final _focusNode = FocusNode();
  late final ChatDetailBloc _chatBloc = Get.find<ChatDetailBloc>();
  late ChatDetailParams params;

  Widget chatDetailBackgroundBuilder(
      Widget child, bool isLoadingList, bool isLoadingDetail) {
    if ((widget.options ?? chatOptions)!.chatDetailBackgroundBuilder != null) {
      return (widget.options ?? chatOptions)!.chatDetailBackgroundBuilder!(
          child, isLoadingList, isLoadingDetail);
    }
    return child;
  }

  @override
  void initState() {
    super.initState();
    params = widget.params ?? getx.Get.arguments;
    _chatBloc.params = params;
    _chatBloc.add(ChatDetailInitialEvent(_scrollController, this));
    chatOptions!.chatOnInitDetail?.call(this);
  }

  @override
  void dispose() {
    _chatBloc.add(ChatDetailDisposeEvent());
    super.dispose();
    _scrollController.dispose();
    _inputController.dispose();
    _focusNode.dispose();
    chatOptions!.chatOnDisposeDetail?.call();
  }

  @override
  Widget build(BuildContext context) {
    Widget child = KeyboardDismissOnTap(
      child: KeyboardVisibilityBuilder(builder: (_, isKeyboardVisible) {
        if (isKeyboardVisible) {
          Timer(const Duration(milliseconds: 500), () {
            if (_scrollController.hasClients) {
              _scrollController.animateTo(
                  _scrollController.position.minScrollExtent,
                  duration: const Duration(milliseconds: 350),
                  curve: Curves.easeIn);
            }
          });
        }
        Widget child = Stack(
          alignment: Alignment.topCenter,
          children: [
            Column(
              children: [
                WidgetProfileChat(
                  params: widget.params,
                  builder: widget.options?.chatDetailProfileBuilder,
                ),
                _buildMessages(),
                Row(
                  children: [
                    const Spacer(),
                    SizedBox(
                      key: _chatBloc.targetAssistantKey,
                      width: 2,
                      height: 2,
                    ),
                  ],
                ),
                WidgetInputChat(
                  builder: widget.options?.chatDetailInputBuilder,
                ),
              ],
            ),
          ],
        );

        return BlocBuilder<ChatListBloc, ChatListState>(
          buildWhen: (previous, current) =>
              previous.isLoading != current.isLoading,
          bloc: Get.find<ChatListBloc>(),
          builder: (_, stateChatList) {
            return BlocBuilder<ChatDetailBloc, ChatDetailState>(
              bloc: Get.find<ChatDetailBloc>(),
              buildWhen: (previous, current) =>
                  previous.isLoading != current.isLoading,
              builder: (_, stateChatDetail) {
                return chatDetailBackgroundBuilder(
                    child, stateChatList.isLoading, stateChatDetail.isLoading);
              },
            );
          },
        );
      }),
    );
    return child;
  }

  Widget _buildMessages() {
    return Expanded(
      child: BlocBuilder<ChatDetailBloc, ChatDetailState>(
        bloc: Get.find<ChatDetailBloc>(),
        builder: (_, state) {
          List<Message>? messages = state.messages;
          if (messages == null || state.isLoading) return const SizedBox();
          return ListView(
            physics: const BouncingScrollPhysics(),
            padding: EdgeInsets.symmetric(horizontal: ChatDetailScreen.padding),
            controller: _scrollController,
            reverse: true,
            children: _getGroupedMessages(messages),
          );
        },
        // ),
      ),
    );
  }

  List<Widget> _getGroupedMessages(List<Message> messages) {
    if (coreConfig.coreThemes.chatOptions!.chatDetailGroupedMessagesBuilder !=
        null) {
      return coreConfig
          .coreThemes.chatOptions!.chatDetailGroupedMessagesBuilder!(messages);
    }
    List<Widget> groupedMessages =
        List.from(<Widget>[const SizedBox(height: 20)]);
    groupBy<Message, String>(messages, (msg) {
      String datetime =
          DateTime.fromMillisecondsSinceEpoch(msg.timestamp!).formatDateTime();
      int otherMessages = 0;
      List<Message> temp = List.from(messages
          .where((e) =>
              DateTime.fromMillisecondsSinceEpoch(e.timestamp!)
                      .formatDateTime() ==
                  datetime &&
              ((msg.doc?.data() ?? msg.tempDoc) as Map)[DB_KEY_FROM] !=
                  ((e.doc?.data() ?? e.tempDoc) as Map)[DB_KEY_FROM] &&
              e.timestamp! > msg.timestamp!)
          .toList())
        ..sorted((a, b) => a.timestamp! - b.timestamp!);
      otherMessages = temp.length;
      return '$datetime-${((msg.doc?.data() ?? msg.tempDoc) as Map)[DB_KEY_FROM]}-$otherMessages';
    }).forEach((when, actualMessages) {
      groupedMessages
          .add(actualMessages.last.widgetMessageOwner(actualMessages));
    });
    return groupedMessages.reversed.toList();
  }
}
