import 'package:_private_core_network/network_resources/resources.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:encrypt/encrypt.dart';
import 'package:get/get.dart';
import 'package:intl/intl.dart';
import 'package:_private_core/_private_core.dart';
import 'package:uuid/uuid.dart';

import '../utils/utils.dart';
import 'chat_detail_screen.dart';

class ChatHelper {
  ChatHelper._();

  //Manage status message sending
  static final Map<String, Future> _messageStatus = <String, Future>{};

  static _getMessageKey(String? peerNo, int? timestamp) => '$peerNo$timestamp';

  static getMessageStatus(String? peerNo, int? timestamp) {
    final key = _getMessageKey(peerNo, timestamp);
    return _messageStatus[key] ?? true;
  }

  static addMessage(String? peerNo, int? timestamp, Future future) {
    final key = _getMessageKey(peerNo, timestamp);
    future.then((_) {
      _messageStatus.remove(key);
    });
    _messageStatus[key] = future;
  }

  //===========
  static Uuid uuid = const Uuid();
  static String generateConversationId() => uuid.v4();

  static final _iv = IV.fromLength(16);
  static String generateEncryptKey() => Key.fromSecureRandom(32).base64;

  static String encrypt(key, input) {
    try {
      return Encrypter(AES(Key.fromBase64(key))).encrypt(input, iv: _iv).base64;
    } catch (e) {
      return input;
    }
  }

  static String decrypt(key, input) {
    try {
      return Encrypter(AES(Key.fromBase64(key)))
          .decrypt(Encrypted.from64(input), iv: _iv);
    } catch (e) {
      return input;
    }
  }

  static findConversationById(
      {required String conversationId,
      required ChatConversationType type}) async {
    QuerySnapshot q = await FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .where(DB_KEY_CONVERSATION_TYPE,
            isEqualTo: type == ChatConversationType.group
                ? CONVERSATION_TYPE_GROUP
                : CONVERSATION_TYPE_SOLO)
        .where(DB_KEY_CONVERSATION_ID, isEqualTo: conversationId)
        .get();
    if (q.size > 0) {
      return q.docs.first.data();
    }
  }

  static findOrCreateSoloConversation(peerId, userType) async {
    QuerySnapshot q = await FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .where(DB_KEY_CONVERSATION_TYPE, isEqualTo: CONVERSATION_TYPE_SOLO)
        .where(DB_KEY_MEMBERS, arrayContains: loggedFirebaseId)
        // .where(DB_KEY_MEMBERS, arrayContains: peerId)
        .get();
    if (q.docs.any(
        (d) => ((d.data() as Map)[DB_KEY_MEMBERS] as List).contains(peerId))) {
      return q.docs
          .firstWhere((d) =>
              ((d.data() as Map)[DB_KEY_MEMBERS] as List).contains(peerId))
          .data();
    } else {
      String id = generateConversationId();
      Map<String, dynamic> data = {
        DB_KEY_CONVERSATION_ID: id,
        DB_KEY_CONVERSATION_TYPE: CONVERSATION_TYPE_SOLO,
        DB_KEY_MEMBERS: [peerId, loggedFirebaseId],
        DB_KEY_LAST_MESSAGE_TIME: 0,
        DB_KEY_ENCRYPT_KEY: generateEncryptKey(),
        DB_KEY_MEMBERS_DETAIL: [
          {
            DB_KEY_FIREBASE_ID: peerId,
            DB_KEY_USER_TYPE: userType,
          },
          {
            DB_KEY_FIREBASE_ID: loggedFirebaseId,
            DB_KEY_USER_TYPE:
                appPrefs.loginUser.isUserVersion ? C_USERS : C_PROS,
          }
        ],
      };
      await FirebaseFirestore.instance
          .collection(C_CONVERSATIONS)
          .doc(id)
          .set(data);
      ContactRepo()
          .addUserOrProToContactListByFirebaseId({"firebaseId": peerId});
      return data;
    }
  }

  static findOrCreateGroupConversation(
      {String? conversationId,
      peers,
      peersDetail,
      groupName,
      thumbnail}) async {
    QuerySnapshot q = await FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .where(DB_KEY_CONVERSATION_TYPE, isEqualTo: CONVERSATION_TYPE_GROUP)
        .where(DB_KEY_CONVERSATION_ID, isEqualTo: conversationId ?? '-1')
        .get();
    if (q.size > 0) {
      return q.docs.first.data();
    } else {
      String id = conversationId ?? generateConversationId();
      Map<String, dynamic> data = {
        DB_KEY_CONVERSATION_ID: id,
        DB_KEY_CONVERSATION_TYPE: CONVERSATION_TYPE_GROUP,
        DB_KEY_GROUP_NAME: groupName,
        DB_KEY_THUMBNAIL: thumbnail,
        DB_KEY_MEMBERS: [...peers, loggedFirebaseId],
        DB_KEY_LAST_MESSAGE_TIME: 0,
        DB_KEY_ENCRYPT_KEY: generateEncryptKey(),
        DB_KEY_MEMBERS_DETAIL: [
          ...peersDetail,
          {
            DB_KEY_FIREBASE_ID: loggedFirebaseId,
            DB_KEY_USER_TYPE:
                appPrefs.loginUser.isUserVersion ? C_USERS : C_PROS,
          }
        ],
      };
      await FirebaseFirestore.instance
          .collection(C_CONVERSATIONS)
          .doc(id)
          .set(data);
      return data;
    }
  }

  static updateConversationInfo(conversationId, data) {
    FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .doc(conversationId)
        .update(data);
  }

  static setLastMessageTime(conversationData, lastTime,
      {alsoMembersDetail = false}) {
    Map<String, dynamic> data = {DB_KEY_LAST_MESSAGE_TIME: lastTime};
    if (alsoMembersDetail) {
      List members = conversationData[DB_KEY_MEMBERS_DETAIL];
      members.firstWhere((e) => e[DB_KEY_FIREBASE_ID] == loggedFirebaseId)[
          DB_KEY_TIMESTAMP] = lastTime + 1;
      data.addAll({DB_KEY_MEMBERS_DETAIL: members});
    }
    FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .doc(conversationData[DB_KEY_CONVERSATION_ID])
        .update(data);
  }

  static setLastTimeTimestamp(conversationData, [timestamp]) async {
    List members = conversationData[DB_KEY_MEMBERS_DETAIL];
    members.firstWhere((e) => e[DB_KEY_FIREBASE_ID] == loggedFirebaseId)[
            DB_KEY_TIMESTAMP] =
        timestamp ??
            DateTime.now()
                .add(const Duration(seconds: 1))
                .millisecondsSinceEpoch;
    await FirebaseFirestore.instance
        .collection(C_CONVERSATIONS)
        .doc(conversationData[DB_KEY_CONVERSATION_ID])
        .update({DB_KEY_MEMBERS_DETAIL: members});
  }

  static removeConversation(id) async {
    FirebaseFirestore.instance.collection(C_CONVERSATIONS).doc(id).delete();
  }

  static getWhen(DateTime date) {
    DateTime now = DateTime.now();
    String when;
    if (date.isToday()) {
      when = 'today'.tr;
    } else if (date.day == now.subtract(const Duration(days: 1)).day) {
      when = 'yesterday'.tr;
    } else {
      when = DateFormat.MMMd().format(date);
    }
    return when;
  }
}
