import 'dart:async';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:get/get.dart';
import 'package:_private_chat/chat_detail/chat_detail_screen.dart';
import 'package:_private_chat/chat_detail/chat_helper.dart';
import 'package:_private_chat/utils/utils.dart';
import 'package:_private_core/_private_core.dart';

import '../../connection_status/bloc/connection_status_bloc.dart';

part 'chat_list_event.dart';
part 'chat_list_state.dart';

String chatGetFullNameUser(data) =>
    data == null ? '' : data['displayName'] ?? data['displayName'] ?? '';

String chatGetThumbnailUser(data) => data == null ? '' : data['thumbnail'];

String chatGetFirebaseIdUser(data) => data == null ? '' : data['firebaseId'];

class ChatListConversation {
  late ChatConversationType conversationType;
  late String conversationId;
  late Map data;
  late List userIds;
  late int unread;
  late Map<String, dynamic> lastMessage;
  bool isEmptyMessages = false;
}

class ChatListBloc extends Bloc<ChatListEvent, ChatListState> {
  StreamSubscription? msgSubscription;

  late final CollectionReference _col =
      FirebaseFirestore.instance.collection(C_CONVERSATIONS);

  ChatListBloc()
      : super(ChatListState(
            conversations: [], searchKeyword: '', usersByKeys: {})) {
    on<ChatListInitEvent>(_onInitial);
    on<ChatListReadItemEvent>(onReadItemEvent);
    on<ChatListDisposeEvent>(onDisposeEvent);
    on<ChatListRemoveEvent>(onRemoveEvent);
    on<ChatListFilterByKeywordEvent>(_onFilter);
    on<ChatListUpdateConversationEvent>(_onUpdateConversation);
  }

  _onUpdateConversation(
      ChatListUpdateConversationEvent event, Emitter<ChatListState> emit) {
    ChatHelper.updateConversationInfo(event.id, event.data);
  }

  _onFilter(ChatListFilterByKeywordEvent event, Emitter<ChatListState> emit) {
    return emit(state.update(
      searchKeyword: event.searchKeyword,
    ));
  }

  onRemoveEvent(ChatListRemoveEvent event, Emitter<ChatListState> emit) {
    ChatHelper.removeConversation(event.conversationId);
    state.conversations
        .removeWhere((e) => e.conversationId == event.conversationId);
    emit(state.update());
  }

  onReadItemEvent(ChatListReadItemEvent event, Emitter<ChatListState> emit) {
    if (event.c != null) {
      if (event.c!.conversationType == ChatConversationType.solo) {
        String firebaseId = event.c!.userIds.first[DB_KEY_FIREBASE_ID];
        Map? user = state.usersByKeys[firebaseId];
        String fullname = chatGetFullNameUser(user);
        String thumbnail = chatGetThumbnailUser(user);
        Get.toNamed(ChatDetailScreen.route,
            arguments: ChatDetailParams(
              event.c!.conversationType,
              peerId: firebaseId,
              peerType: event.detailType ??
                  ((user?['uniqueId']?.startsWith('U') ?? false)
                      ? ChatDetailType.users
                      : ChatDetailType.pros),
              name: fullname,
              thumbnail: thumbnail,
            ));
      } else {
        Get.toNamed(
          ChatDetailScreen.route,
          arguments: ChatDetailParams(
            event.c!.conversationType,
            conversationId: event.c!.conversationId,
            name: event.c!.data[DB_KEY_GROUP_NAME],
            thumbnail: event.c!.data[DB_KEY_THUMBNAIL],
          ),
        );
      }
    } else if (event.firebaseId != null) {
      Get.toNamed(
        ChatDetailScreen.route,
        arguments: ChatDetailParams(
          ChatConversationType.solo,
          peerId: event.firebaseId,
          peerType: event.detailType,
          name: event.fullName,
          thumbnail: event.thumbnail,
        ),
      );
    }
  }

  onDisposeEvent(event, Emitter<ChatListState> emit) {
    msgSubscription?.cancel();
  }

  _onInitial(ChatListInitEvent event, Emitter<ChatListState> emit) async {
    emit(state.update(isLoading: true));
    QuerySnapshot q = await _col
        .where(DB_KEY_MEMBERS, arrayContains: loggedFirebaseId)
        .orderBy(DB_KEY_LAST_MESSAGE_TIME, descending: true)
        .get();
    List requestUsers = [];
    List<ChatListConversation> cons = [];
    int totalUnread = 0;
    for (var d in q.docs) {
      Map docData = d.data() as Map;
      int lastTime = docData[DB_KEY_MEMBERS_DETAIL].firstWhere((e) =>
              e[DB_KEY_FIREBASE_ID] == loggedFirebaseId)[DB_KEY_TIMESTAMP] ??
          0;
      ChatListConversation conversation = ChatListConversation()
        ..conversationId = d.id
        ..userIds = docData[DB_KEY_MEMBERS_DETAIL]
            .where((e) => e[DB_KEY_FIREBASE_ID] != loggedFirebaseId)
            .toList()
        ..conversationType = ChatConversationType.values
            .byName(docData[DB_KEY_CONVERSATION_TYPE])
        ..data = docData;
      requestUsers.addAll(docData[DB_KEY_MEMBERS_DETAIL].toList());

      if (docData[DB_KEY_LAST_MESSAGE_TIME] != null &&
          docData[DB_KEY_LAST_MESSAGE_TIME] != 0) {
        QuerySnapshot query = await _col
            .doc(d.id)
            .collection(C_MESSAGES)
            .where(DB_KEY_TIMESTAMP, isGreaterThan: lastTime)
            .get();
        int unread = query.size;
        totalUnread += unread;
        conversation.unread = unread;

        //Last message
        DocumentSnapshot docLastMessage = await _col
            .doc(d.id)
            .collection(C_MESSAGES)
            .doc('${docData[DB_KEY_LAST_MESSAGE_TIME]}')
            .get();
        if (docLastMessage.data() != null) {
          conversation.lastMessage =
              docLastMessage.data() as Map<String, dynamic>;
          conversation.lastMessage[DB_KEY_CONTENT] = ChatHelper.decrypt(
              conversation.data[DB_KEY_ENCRYPT_KEY],
              conversation.lastMessage[DB_KEY_CONTENT]);
        } else {
          conversation.lastMessage = {};
        }

        cons.add(conversation);
      } else {
        conversation.unread = 0;
        conversation.isEmptyMessages = true;
        conversation.lastMessage = {};
        cons.add(conversation);
      }
    }

    Map usersByKeys = await _getUsersByKeys(requestUsers);
    _getConnectionStatusUsers(requestUsers);

    emit(state.update(
      isLoading: false,
      conversations: cons,
      unread: totalUnread,
      usersByKeys: usersByKeys,
    ));

    msgSubscription ??= _col
        .where(DB_KEY_MEMBERS, arrayContains: loggedFirebaseId)
        .snapshots()
        .listen((docs) {
      // add(ChatListInitEvent());
      appDebugPrint('state.conversations: ${state.conversations}');
      appDebugPrint('listen.conversations: ${docs.docs.length}');
      if (docs.docs.length != state.conversations.length) {
        add(ChatListInitEvent());
      }
    });
  }

  void _getConnectionStatusUsers(List requestUsers) async {
    List<String> users =
        requestUsers.map((e) => e[DB_KEY_FIREBASE_ID].toString()).toList();
    Get.find<ConnectionStatusBloc>().add(ConnectionStatusInitEvent(users));
  }

  Future<Map> _getUsersByKeys(requestUsers) async {
    Map usersByKeys = state.usersByKeys;
    for (var e in requestUsers) {
      if (state.usersByKeys[e[DB_KEY_FIREBASE_ID]] == null) {
        Map<String, dynamic>? data = (await FirebaseFirestore.instance
                .collection(e[DB_KEY_USER_TYPE].endsWith('s')
                    ? e[DB_KEY_USER_TYPE]
                    : '${e[DB_KEY_USER_TYPE]}s')
                .doc(e[DB_KEY_FIREBASE_ID])
                .get())
            .data();
        usersByKeys[e[DB_KEY_FIREBASE_ID]] = data;
      }
    }
    return usersByKeys;
  }
}
